;; Generated package description from window-commander.el  -*- no-byte-compile: t -*-
(define-package "window-commander" "3.0.0.20230628.182959" "Simply execute commands on windows" '((emacs "27.1")) :commit "dc7f6f264e67ec4d4eeca253078bdd9a9a136e67" :authors '(("Daniel Semyonov" . "daniel@dsemy.com")) :maintainer '("Daniel Semyonov" . "daniel@dsemy.com") :keywords '("convenience") :url "https://dsemy.com/projects/window-commander")
